"""Utilities for loading SatelliteModem subclasses.
"""

import logging
import importlib.util
import inspect
import os
import tempfile
import shutil
import subprocess
from typing import Type
from pathlib import Path

from . import modems
from .modem import SatelliteModem
from .common import ModemModel

_log = logging.getLogger(__name__)

GITHUB_TOKEN = os.getenv('GITHUB_TOKEN')
GITHUB_ORG = 'inmarsat-enterprise'
GITHUB_REPOS = [
    'quectel-cc200a',
    'skywave-st2-ogx',
    'skywave-st2-idp',
]


# cache: {Path: loaded_module}
_module_cache: dict[Path, object] = {}


def load_module_from_path(module_path: Path):
    """Load a Python module directly from a file path.
    
    Avoid touching sys.path.
    """
    if module_path in _module_cache:
        return _module_cache[module_path]
    module_name = module_path.stem
    spec = importlib.util.spec_from_file_location(module_name, module_path)
    if spec is None or spec.loader is None:
        raise ImportError(f'Cannot load module from {module_path}')
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    _module_cache[module_path] = module
    return module


def clone_and_load_modem_classes(repo_urls: list[str],
                                 branch: str = 'main',
                                 download_path: str = '',
                                 ) -> dict[str, Type[SatelliteModem]]:
    """Clone multiple Git repositories and load subclasses of SatelliteModem.

    Args:
        repo_urls (list[str]): A list of Git repository URLs.
        branch (str): The branch to clone. Defaults to 'main'.

    Returns:
         A dictionary of modem class names and their corresponding classes.
    """
    modem_classes: dict[str, Type[SatelliteModem]] = {}
    # Create a temporary directory to clone repositories
    with tempfile.TemporaryDirectory() as temp_dir:
        for repo_url in repo_urls:
            repo_name = repo_url.split('/')[-1].replace('.git', '')
            repo_path = os.path.join(temp_dir, repo_name)
            _log.debug('Cloning git repository into %s...', repo_path)
            result = subprocess.run(
                ['git', 'clone', '--depth', '1', '--branch',
                 branch, repo_url, repo_path],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            if result.returncode != 0:
                _log.error("Failed to clone repository %s: %s",
                           repo_url, result.stderr)
                continue
            _log.debug("Git repository %s cloned successfully.", repo_name)
            # Find Python files in the repository and load modem classes
            for root, _, files in os.walk(repo_path):
                if any(p in Path(root).parts for p in ['tests', 'examples']):
                    continue
                for file in files:
                    if file in ['__init__.py', 'main.py']:
                        continue
                    if file.endswith('.py'):
                        file_path = Path(root) / file
                        try:
                            submodule = load_module_from_path(file_path)
                        except Exception as e:
                            _log.exception('Failed to load module %s: %s',
                                           file_path, e)
                            continue
                        for _, cls in inspect.getmembers(submodule, inspect.isclass):
                            if (issubclass(cls, SatelliteModem) and
                                cls is not SatelliteModem):
                                modem_classes[cls.__name__] = cls
                                _log.debug('Loaded modem class: %s', cls.__name__)
                        # Copy to download_path if requested
                        if download_path:
                            os.makedirs(download_path, exist_ok=True)
                            dest_path = Path(download_path) / file
                            shutil.copy(file_path, dest_path)
                            _log.debug('Copied %s to %s', file, dest_path)
    return modem_classes


def mutate_modem(modem: SatelliteModem, **kwargs) -> SatelliteModem:
    """Mutate and return the model-specific subclass of the satellite modem.
    
    Attempts to find the module in `modems_path` which defaults to the `modems`
    folder under `pynanomodem`.
    If not found, will attempt to clone/download from GitHub private repository
    if a GITHUB_TOKEN environment variable is present.
    
    Args:
        modem (SatelliteModem): The base/unknown modem.
        **module (module): The module containing the subclass python files.
            Downloaded files from GitHub will be stored here.
        **mixin (SatelliteModem): Optional mixin extension subclass to apply.
    
    Returns:
        Subclass of SatelliteModem.
    
    Raises:
        ModuleNotFoundError if unable to load the subclass.
    """
    was_connected = modem.is_connected()
    if not was_connected:
        modem.connect()
    model = modem.get_model()
    if model == ModemModel.UNKNOWN:
        raise ModuleNotFoundError('Unrecognized modem')
    if model == modem._model:
        return modem
    pymodule = kwargs.pop('module', modems)
    modems_path = Path(pymodule.__path__[0])
    file_tag = f'{model.name.lower()}.py'
    modem_path = next(
        (p for p in modems_path.glob('*.py') if p.name.endswith(file_tag)),
        None,
    )
    if modem_path is None:
        try:
            token = kwargs.get('github_token', GITHUB_TOKEN)
            if token:
                _log.debug('Attempting to clone model-specific subclass...')
                org = kwargs.get('github_org_name', GITHUB_ORG)
                repos: list[str] = kwargs.get('github_repos', GITHUB_REPOS)
                for repo_name in repos:
                    if repo_name.replace('-', '_').endswith(model.name.lower()):
                        _log.info('Copying %s from GitHub to %s',
                                    model.name, modems_path)
                        repo_url = (f'https://{token}@github.com'
                                    f'/{org}/pynanomodem-{repo_name}')
                        clone_and_load_modem_classes(
                            [repo_url], download_path=str(modems_path)
                        )
                # refresh after download
                modem_path = next(
                    (p for p in modems_path.glob('*.py') if p.name.endswith(file_tag)),
                    None,
                )
        except Exception as e:
            raise ModuleNotFoundError(f'No module for {model.name}') from e
    # Check if download still did not get the target file
    if modem_path is None:
        raise ModuleNotFoundError(f'No Python file found for {model.name}')
    submodule = load_module_from_path(modem_path)
    for _, candidate in inspect.getmembers(submodule, inspect.isclass):
        if issubclass(candidate, SatelliteModem):
            if getattr(candidate, '_model', None) == model:
                # Disconnect for state prior to mutation
                if not was_connected:
                    modem.disconnect()
                mixin = kwargs.get('mixin')
                if mixin and issubclass(mixin, SatelliteModem):
                    Extended = type(
                        f'Extended{candidate.__name__}',
                        (candidate, mixin),
                        {}
                    )
                    modem.__class__ = Extended
                else:
                    modem.__class__ = candidate
                modem._post_mutate()
                return modem
    # Fall-through
    raise ModuleNotFoundError(f'No subclass found for {model.name}')



"""Module for communicating with Viasat IoT Nano modems."""

from pyatcommand import AtClient, AtTimeout

from .common import (
    BeamStateIdp,
    BeamType,
    BeamTypeIdp,
    BeamTypeOgx,
    ControlStateIdp,
    DataFormat,
    EventNotification,
    GeoBeam,
    GnssMode,
    GnssModeQuectel,
    GnssModeSkywave,
    LastErrorCode,
    MessagePriorityIdp,
    MessageState,
    MessageStateIdp,
    MessageStateOgx,
    MessageTypeOgx,
    ModemManufacturer,
    ModemModel,
    NetInfo,
    NetworkProtocol,
    NetworkState,
    NetworkStateOgx,
    OperatingMode,
    PowerMode,
    SatelliteId,
    ServiceClassOgx,
    SignalLevelIdp,
    SignalQuality,
    WakeupInterval,
    WakeupIntervalIdp,
    WakeupIntervalOgx,
)
from .loader import clone_and_load_modem_classes, mutate_modem
from .location import GnssFixQuality, GnssFixType, GnssLocation, GnssSatelliteInfo
from .message import IotNanoMessage, MoMessage, MtMessage
from .modem import SatelliteModem

__all__ = [
    'AtClient',
    'AtTimeout',
    'LastErrorCode',
    'SatelliteModem',
    'ModemManufacturer',
    'ModemModel',
    'BeamType',
    'BeamTypeIdp',
    'BeamStateIdp',
    'BeamTypeOgx',
    'IotNanoMessage',
    'MessageState',
    'MessageStateIdp',
    'MessageStateOgx',
    'MoMessage',
    'MtMessage',
    'NetworkProtocol',
    'ServiceClassOgx',
    'NetworkState',
    'SignalQuality',
    'NetInfo',
    'DataFormat',
    'EventNotification',
    'WakeupInterval',
    'WakeupIntervalIdp',
    'WakeupIntervalOgx',
    'PowerMode',
    'GnssMode',
    'GnssLocation',
    'GnssFixType',
    'GnssFixQuality',
    'GnssSatelliteInfo',
    'OperatingMode',
    'ControlStateIdp',
    'NetworkStateOgx',
    'MessagePriorityIdp',
    'MessageTypeOgx',
    'GnssModeSkywave',
    'GnssModeQuectel',
    'SignalLevelIdp',
    'GeoBeam',
    'SatelliteId',
    'clone_and_load_modem_classes',
    'mutate_modem',
]



"""IoT Nano modem abstract base class implementation.

Does not enforce any @abstractmethod, allowing a generic default class to
provide basic methods to query any modem.
"""

import logging
from abc import ABC
from typing import Optional, Union

from pyatcommand import AtClient, xmodem_bytes_handler

from .common import (
    ModemManufacturer,
    ModemModel,
    NetworkProtocol,
    NetworkState,
    NetInfo,
    SignalQuality,
    PowerMode,
    WakeupInterval,
    GnssMode,
    EventNotification,
)
from .location import GnssLocation
from .message import MoMessage, MtMessage

_log = logging.getLogger(__name__)


class SatelliteModem(AtClient, ABC):
    """Abstract Base Class for a IoT Nano modem."""
    
    _manufacturer: ModemManufacturer = ModemManufacturer.UNKNOWN
    _model: ModemModel = ModemModel.UNKNOWN
    _network: NetworkProtocol = NetworkProtocol.IDP
    
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._mobile_id: str = ''
        self._firmware_version: str = ''
        self._command_timeout = 1
    
    def _post_mutate(self) -> None:
        """Run after mutating the base class to apply __init__ settings."""
    
    def connect(self, **kwargs) -> None:
        super().connect(**kwargs)
        self._autoconfig = False   # avoid potential false V0/E0 detection
    
    def disconnect(self) -> None:
        super().disconnect()
        self._mobile_id = ''
        self._firmware_version = ''
        self._autoconfig = True   # allow for new modem detection
    
    @property
    def network(self) -> NetworkProtocol:
        return self._network
    
    @property
    def manufacturer(self) -> str:
        """Get the manufacturer name."""
        return self._manufacturer.name
    
    @property
    def model(self) -> str:
        """Get the modem model."""
        return self._model.name
    
    def get_model(self) -> ModemModel:
        """Check the model of the connected modem."""
        model = self._model
        mfr_res = self.send_command('ATI')
        if mfr_res.ok and mfr_res.info:
            if 'ORBCOMM' in mfr_res.info.upper():
                model_res = self.send_command('ATI4')
                if model_res.ok and model_res.info:
                    if 'ST2' in model_res.info:
                        proto_res = self.send_command('ATI5')
                        if proto_res.ok and proto_res.info:
                            if proto_res.info == '8':
                                model = ModemModel.ST2_IDP
                            elif proto_res.info == '10':
                                model = ModemModel.ST2_OGX
                            else:
                                raise ValueError('Unsupported protocol value')
            elif 'QUECTEL' in mfr_res.info.upper():
                model = ModemModel.CC200A
        if self._model != ModemModel.UNKNOWN and model != self._model:
            _log.warning('Detected %s but expected %s', model.name, self.model)
        elif model == ModemModel.UNKNOWN:
            _log.warning('Unable to determine modem model')
        return model
        
    @property
    def firmware_version(self) -> str:
        """Get the modem firmware version."""
        if not self._firmware_version:
            resp = self.send_command('AT+GMR', prefix='+GMR:')
            if resp.ok and resp.info:
                # TODO: parse components
                self._firmware_version = resp.info
        return self._firmware_version
    
    @property
    def mobile_id(self) -> str:
        """The modem's globally unique identifier."""
        if self._mobile_id:
            return self._mobile_id
        resp = self.send_command('AT+GSN', prefix='+GSN:')
        if resp.ok and resp.info:
            return resp.info
        return ''
    
    def initialize(self, **kwargs) -> bool:
        """Initialize settings for the modem.
        
        Settings for event notifications/monitoring and URC notifications.
        """
        return True
    
    # @abstractmethod
    def get_network_state(self) -> NetworkState:
        """Get the current network state."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_netinfo(self) -> NetInfo:
        """Get details of the acquisition process."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_snr(self) -> float:
        """Get the SNR (C/N0) of the modem."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_signal_quality(self, **kwargs) -> SignalQuality:
        """Get the qualitative value of the signal."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    def is_transmit_allowed(self) -> bool:
        """Check if message transmission is allowed."""
        return self.get_network_state().is_registered()
    
    def is_blocked(self) -> bool:
        """Check if line of sight is blocked."""
        return self.get_network_state().is_blocked()
    
    def is_updating_network(self) -> bool:
        """Check if the modem is updating its network configuration.
        
        The modem should not be powered down during a network update.
        """
        return self.get_network_state().is_updating()
    
    def is_muted(self) -> bool:
        """Check if modem transmission is muted."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def mo_message_send(self, data: bytes, **kwargs) -> Union[MoMessage, None]:
        """Send a mobile-originated message.
        
        Args:
            data (bytes): The binary data to send (including SIN/MIN bytes).
            **message_id (int|str): A unique message identifier in the Tx queue.
                If not provided, one will be assigned.
        
        Returns:
            `MoMessage` (including unique ID) if successful.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def mo_message_cancel(self, message_id: Union[str, int]) -> bool:
        """Cancel a previously submitted message.
        
        Args:
            message (str): The unique identifier in the Tx queue
                to attempt to cancel.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_mo_message_queue(self, message: Optional[MoMessage] = None) -> list[MoMessage]:
        """Get mobile-originated message(s) in the Tx queue.
        
        Args:
            message (MoMessage): If present, only retrieve the specified status.
        
        Returns:
            A list of `MoMessage`(s) in the Tx queue. If `message` was
                specified and is queued, it will be the only list element.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    def clear_mo_message_queue(self) -> bool:
        """Iterate the modem Tx queue to remove completed messages."""
        tx_queue = self.get_mo_message_queue()
        success = True
        for msg in tx_queue:
            if not self.mo_message_delete(msg.id):                              # type: ignore
                success = False
        return success
    
    def mo_message_delete(self, message_id: Union[str, int]) -> bool:
        """Remove a completed mobile-originated message from the Tx queue.
        
        Args:
            message_id (str): The unique identifier in the Tx queue to delete.
        
        Returns:
            True if successful.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_mt_message_queue(self, message: Optional[MtMessage] = None) -> list[MtMessage]:
        """Get mobile-terminated message(s) in the Rx queue.
        
        Args:
            message (MtMessage): If present, only get the specified status.
        
        Returns:
            A list of queued `MtMessage`(s). If `message` was specified and
                is queued, it will be the only list element.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    def clear_mt_message_queue(self) -> bool:
        """Iterate the modem Rx queue to remove completed messages."""
        rx_queue = self.get_mt_message_queue()
        success = True
        for msg in rx_queue:
            if not self.mt_message_delete(msg.id):                              # type: ignore
                success = False
        return success
    
    # @abstractmethod
    def mt_message_recv(self, message: MtMessage, **kwargs) -> Union[MtMessage, None]:
        """Retrieve the specified message (by name) from the Rx queue.
        
        Args:
            message (MtMessage): The message with `name` specified.
        
        Returns:
            The message content and meta if it was in the Rx queue, or None.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    def mt_message_delete(self, message_id: Union[str, int]) -> bool:
        """Remove a completed mobile-terminated message from the Rx queue.
        
        Args:
            message_id (str): The unique identifier in the Rx queue to delete.
        
        Returns:
            True if successful.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    def send_bytes_data_mode(self, data: bytes, **kwargs) -> Union[int, None]:
        """Send using XMODEM protocol.
        
        Args:
            data (bytes): The data to send.
            **callback (Callable): Optional callback to validate byte count.
        
        Returns:
            The number of bytes sent if no callback is specified otherwise None.
        """
        self.data_mode = True
        success = xmodem_bytes_handler(self._serial, 'send', data)              # type: ignore
        self.data_mode = False
        callback = kwargs.get('callback')
        if callable(callback):
            callback(len(data) if success else 0)
            return None
        return len(data) if success else 0
        
    def recv_bytes_data_mode(self, **kwargs) -> Union[bytes, None]:
        """Receive data using XMODEM protocol.
        
        Args:
            **callback (Callable): Callback to receive the message data.
        
        Returns:
            If `callback` is not specified returns the bytes, otherwise None. 
        """
        self.data_mode = True
        data: bytes = xmodem_bytes_handler(self._serial, 'recv', None)          # type: ignore
        self.data_mode = False
        data = data.rstrip(b'\x1a') or b''
        callback = kwargs.get('callback')
        if callable(callback):
            callback(data)
            return None
        return data
    
    def get_gnss_mode(self) -> GnssMode:
        """Get the GNSS operating mode for which systems are in use."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    def set_gnss_mode(self, gnss_mode: Union[GnssMode, int], **kwargs) -> bool:
        """Set the GNSS operating mode for which systems to use."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    def get_gnss_interval(self) -> int:
        """Get the GNSS refresh interval in seconds."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    def set_gnss_interval(self, gnss_interval: int) -> bool:
        """Set the GNSS refresh interval in seconds.
        
        Args:
            gnss_interval (int): The number of seconds between GNSS updates.
                Must be in range 0..30. 0 disables refresh.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_location(self, **kwargs) -> Union[GnssLocation, None]:
        """Get the modem location.
        
        Args:
            **stale_secs (int): The maximum age of the fix to return.
            **wait_secs (int): The maximum time to wait for a fix.
            **nmea_sentences (str): A CSV list of valid sentence types.
                May include RMC,GGA,GSA,GSV.
                
        Returns:
            GnssLocation or None if a GNSS timeout occurs.
        """
        # stale_secs = kwargs.get('stale_secs', 1)
        # wait_secs = kwargs.get('wait_secs', 45)
        # nmea_sentences = kwargs.get('nmea_sentences')
        # if not isinstance(nmea_sentences, str):
        #     nmea_sentences = 'RMC,GGA,GSA'
        # nmea_list = nmea_sentences.replace('"', '').split(',')
        # nmea_sentences = ','.join(f'"{x.strip()}"' for x in nmea_list)
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_system_time(self) -> int:
        """Get UTC epoch seconds from the modem."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_wakeup_interval(self) -> WakeupInterval:
        """Get the modem wakeup interval."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def set_wakeup_interval(self,
                            wakeup_interval: Union[WakeupInterval, int],
                            **kwargs) -> bool:
        """Request a modem wakeup interval.
        
        The wakeup interval will not actually be changed until acknowledged
        by the network.
        
        Args:
            wakeup_interval (WakeupInterval): The desired wakeup interval.
        
        Returns:
            True if accepted. Note that the change only happens when
                acknowledged by the network.
        """
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def get_power_mode(self) -> PowerMode:
        """Get the power mode setting of the modem."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    # @abstractmethod
    def set_power_mode(self, power_mode: Union[PowerMode, int]) -> bool:
        """Set the power mode of the modem.
        
        Args:
            power_mode (PowerMode): The desired power mode.
        
        Returns:
            True if successful.
        """
        raise NotImplementedError('Implement in model-specific subclass')

    def set_event_mask(self, events_bitmask: int, **kwargs) -> bool:
        """Configure modem events that trigger notifications.
        
        Args:
            events_bitmask (int): The bitmask of events to monitor.
        """
        if not isinstance(events_bitmask, int):
            raise ValueError('Invalid bitmask')
        raise NotImplementedError('Implement in model-specific subclass')
    
    def get_event_mask(self) -> int:
        """Read the modem event notification configuration bitmask."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    def get_active_events_mask(self) -> int:
        """Get the active events."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    def set_monitor_network_trace(self, enable: bool = True) -> bool:
        """Enable or disable monitoring of network trace events."""
        if self.network == NetworkProtocol.IDP:
            return self.set_trace_events_monitor([(3, 1)])
        return True   # built into %NETINFO
        
    def get_trace_events_monitor(self) -> list[tuple[int, int]]:
        """Get the monitored trace events.
        
        Returns a list of (class, subclass) tuples.
        """
        events: list[tuple[int, int]] = []
        resp = self.send_command('AT%EVMON', prefix='%EVMON:')
        if resp.ok and resp.info:
            events_str = [e.strip() for e in resp.info.split(',')]
            for event in events_str:
                events.append(tuple([int(v) for v in event.split('.')]))        # type: ignore
        else:
            _log.warning('Unable to determine monitored trace events')
        return events
    
    def set_trace_events_monitor(self, event_list: list[tuple[int,int]]) -> bool:
        """Set the monitored trace events.
        
        Args:
            event_list (list): A list of (class, subclass) tuples to monitor.
        """
        if not isinstance(event_list, list):
            raise ValueError('Invalid event list')
        for e in event_list:
            if (not (isinstance(e, tuple) and
                     len(e) == 2 and
                     all(isinstance(i, int) for i in e))):
                raise ValueError('Invalid event list')
        events_str = ','.join(f'{e[0]}.{e[1]}' for e in event_list)
        return self.send_command(f'AT%EVMON={events_str}').ok
    
    # @abstractmethod
    def get_last_error_code(self) -> int:
        """Get the last numeric error code result from the modem."""
        resp = self.send_command('ATS80?')
        if resp.ok and resp.info:
            return int(resp.info)
        else:
            _log.warning('Unable to determine last error code')
            return 0
    
    def get_urc_event(self, urc: str) -> Union[EventNotification, None]:
        """Parse a URC to derive an event notification."""
        raise NotImplementedError('Implement in model-specific subclass')
    
    def report_debug(self, **kwargs):
        """Log a debug report."""
        raise NotImplementedError('Implement in model-specific subclass')



"""Common constants, classes and methods."""

import json
from dataclasses import dataclass
from enum import Enum, IntEnum, IntFlag
from typing import Optional, Union


class LastErrorCode(IntEnum):
    """Error codes from `get_last_error_code`."""
    NONE = 0
    INVALID_CMD_CRC = 100
    UNKNOWN_CMD = 101
    INVALID_CMD_PARAMETER = 102
    MSG_TOO_LARGE = 103
    DATA_MODE_ERROR = 104
    SYSTEM_ERROR = 105
    QUEUES_FULL = 106
    MSG_NAME_ALREADY_IN_USE = 107
    GNSS_TIMEOUT = 108
    MSG_UNAVAILABLE = 109
    RESOURCE_BUSY = 111
    READ_ONLY_PARAMETER = 112
    GNSS_FIX_PENDING = 113
    MSG_NOT_FOUND = 120
    INVALID_CLASS_OF_SERVICE = 121
    INVALID_LIFETIME = 122
    INVALID_MSG_LENGTH = 123
    DATA_FORMAT_ERROR = 124
    INVALID_BLOCKID = 125
    MSG_UNABLE_TO_CANCEL = 126
    MSG_UNABLE_TO_PAUSE = 127
    MSG_UNABLE_TO_DELETE = 128
    MSG_MO_PAUSED = 129
    MSG_UNABLE_TO_RESUME = 130
    INVALID_EVENT_ID = 131
    NO_EVENT_DATA = 132
    NETWORK_TX_NOT_READY = 133
    INFO_UNAVAILABLE = 134


class ModemManufacturer(IntEnum):
    """Recognized modem/module manufacturers."""
    UNKNOWN = 0
    SKYWAVE = 1
    HONEYWELL = 2
    QUECTEL = 3
    UBLOX = 4


class ModemModel(IntEnum):
    """Recognized modem/module models."""
    UNKNOWN = 0
    OGI = 1
    ST2_IDP = 2
    CC200A = 3
    ST2_OGX = 4


class NetworkProtocol(IntEnum):
    """Network protocol IDP/OGx."""
    IDP = 0
    OGX = 1


class DataFormat(IntEnum):
    """Data format used for sending or receiving messages via AT command."""
    BIN = 0
    TEXT = 1
    HEX = 2
    BASE64 = 3


class NetworkState(IntEnum):
    """The network state."""
    def is_awaiting_gnss(self) -> bool:
        return self.name in ['GNSS_WAIT']
    
    def is_searching(self) -> bool:
        return self.name.startswith('BEAM')
    
    def is_updating(self) -> bool:
        return self.name in ['BB_DOWNLOAD', 'UPDATING']
    
    def is_registering(self) -> bool:
        return self.name == 'REGISTERING'
    
    def is_registered(self) -> bool:
        return self.name in ['ACTIVE', 'CONNECTED']
    
    def is_blocked(self) -> bool:
        return self.name in ['BLOCKED']


class ControlStateIdp(NetworkState):
    """States of the IDP modem internal network acquisition process."""
    STOPPED = 0
    GNSS_WAIT = 1
    SEARCH_START = 2
    BEAM_SEARCH = 3
    BEAM_FOUND = 4
    BEAM_ACQUIRED = 5
    BEAM_SWITCH = 6
    REGISTERING = 7
    RECEIVE_ONLY = 8    
    BB_DOWNLOAD = 9
    ACTIVE = 10
    BLOCKED = 11
    CONFIRM_PREVIOUS_BEAM = 12
    CONFIRM_REQUESTED_BEAM = 13
    CONNECT_CONFIRMED_BEAM = 14


class NetworkStateOgx(NetworkState):
    """Network state used by OGx."""
    STOPPED = 0
    GNSS_WAIT = 1
    BEAM_SEARCH = 2
    UPDATING = 3
    REGISTERING = 4
    CONFIRMING = 5
    CONNECTED = 6


class BeamType(IntEnum):
    """The beam type used for signal analysis."""
    def is_global(self) -> bool:
        return self.name in ['GB', 'SAM_GB']
    
    def is_nominal(self) -> bool:
        return 'NOM' in self.name or self.name == 'RB'


class BeamTypeIdp(BeamType):
    """The beam type used for IDP."""
    GB = 0
    RB = 1


class BeamTypeOgx(BeamType):
    """The beam type used for OGx."""
    ILC_NOM_FL = 1
    ILC_NOM_RL = 2
    SAM_GB = 3
    SAM_RB = 4
    ILC_CFG = 5
    ILC_REG = 6


class BeamStateIdp(IntEnum):
    """The IDP beam search state."""
    IDLE = 0
    SEARCH_TRAFFIC_ANY = 1
    SEARCH_TRAFFIC_LAST = 2
    SEARCH_TRAFFIC_NEXT = 4
    SEARCH_BB = 5
    DELAY = 6


class MessageState(IntEnum):
    """Base class for IDP and OGx message states."""
    def is_complete(self):
        return self.name in ['RX_COMPLETE', 'TX_COMPLETE', 'TX_FAIL',
                             'TX_ABORT', 'TX_EXPIRED', 'TX_CANCELLED']
    
    def is_success(self):
        return self.name in ['RX_COMPLETE', 'TX_COMPLETE']
    
    def is_failed(self):
        return self.name in ['TX_FAIL']
    
    def is_expired(self):
        return self.name in ['TX_ABORT', 'TX_EXPIRED']
    
    def is_cancelled(self):
        return self.name in ['TX_CANCELLED']


class MessageStateIdp(MessageState):
    """The IDP message state."""
    UNAVAILABLE = 0
    RX_COMPLETE = 2
    RX_RETRIEVED = 3
    TX_READY = 4
    TX_SENDING = 5
    TX_COMPLETE = 6
    TX_FAIL = 7
    TX_ABORT = 8


class MessageStateOgx(MessageState):
    """The OGx message state."""
    TX_INITIALIZING = 1
    TX_OFFLINE = 2
    TX_READY = 3
    TX_SENDING = 4
    RX_COMPLETE = 5
    TX_COMPLETE = 6
    TX_FAIL = 7
    TX_EXPIRED = 8
    TX_CANCELLED = 14


class MessagePriorityIdp(IntEnum):
    """The IDP message priority.
    
    Weights selection of MO fragments to send. This value is not transported
    over-the-air or indicated on the Messaging API.
    """
    NONE = 0   # Used for IDP MT messages only
    HIGH = 1
    NORMAL = 2
    LOW = 3
    LOWEST = 4


class ServiceClassOgx(IntEnum):
    """The OGx service class.
    
    Weights selection of message fragments to send. This value is indicated
    on the Messaging API.
    """
    PREMIUM = 1
    NORMAL = 2
    BACKGROUND = 3


class MessageTypeOgx(IntEnum):
    """The type of message for OGx."""
    SMALL = 1
    SINGLE_BLOCK = 2   # 1024..16000 bytes


class OperatingMode(IntEnum):
    """The operating mode as interpreted by the network API."""
    ON_OFF = 0
    WAKEUP = 1
    ROS = 2
    HYBRID = 3


class GnssMode(IntEnum):
    """The GNSS operating mode."""


class GnssModeSkywave(GnssMode):
    """GNSS mode mappings for SkyWave/Orbcomm."""
    GPS = 0   # default
    GLONASS = 1
    BEIDOU = 2
    GALILEO  = 3
    GPS_GLONASS = 10
    GPS_BEIDOU = 11
    GLONASS_BEIDOU = 12
    GPS_GALILEO = 13
    GLONASS_GALILEO = 14
    BEIDOU_GALILEO = 15


class GnssModeQuectel(GnssMode):
    """GNSS mode mappings for Quectel."""
    GPS = 0
    GPS_BEIDOU = 1
    GPS_GLONASS = 2
    GPS_GALILEO = 3
    GPS_GLONASS_GALILEO_BEIDOU = 4


class WakeupInterval(IntEnum):
    """Base class for IDP or OGx wakeup intervals."""
    def seconds(self) -> int:
        if self.name == 'NONE':
            return 5
        units, value = self.name.split('_')
        if units == 'SECONDS':
            return int(value)
        if units == 'MINUTES':
            return int(value) * 60
        return int(value) * 3600
    
    @classmethod
    def nearest(cls, seconds: int) -> 'WakeupInterval':
        """Get the nearest interval to the specified number of seconds."""
        raise NotImplementedError('Must be called on subclass')


class WakeupIntervalIdp(WakeupInterval):
    """The modem wakeup interval for IDP."""
    NONE = 0
    SECONDS_30 = 1
    MINUTES_1 = 2
    MINUTES_3 = 3
    MINUTES_10 = 4
    MINUTES_30 = 5
    HOURS_1 = 6
    MINUTES_2 = 7
    MINUTES_5 = 8
    MINUTES_15 = 9
    MINUTES_20 = 10
    
    @classmethod
    def nearest(cls, seconds: int) -> 'WakeupIntervalIdp':
        if seconds == 0:
            return WakeupIntervalIdp.NONE
        if seconds <= 45:
            return WakeupIntervalIdp.SECONDS_30
        if seconds < 3600:
            minutes = round(seconds / 60, 0)
            if minutes < 2:
                return WakeupIntervalIdp.MINUTES_1
            if minutes < 3:
                return WakeupIntervalIdp.MINUTES_2
            if minutes < 5:
                return WakeupIntervalIdp.MINUTES_3
            if minutes < 10:
                return WakeupIntervalIdp.MINUTES_5
            if minutes < 20:
                return WakeupIntervalIdp.MINUTES_15
            if minutes < 30:
                return WakeupIntervalIdp.MINUTES_20
            if minutes < 45:
                return WakeupIntervalIdp.MINUTES_30
        return WakeupIntervalIdp.HOURS_1


class WakeupIntervalOgx(WakeupInterval):
    """The modem wakeup interval for OGx."""
    NONE = 0
    SECONDS_30 = 1
    MINUTES_1 = 2
    MINUTES_2 = 3
    MINUTES_3 = 4
    MINUTES_5 = 5
    MINUTES_10 = 6
    MINUTES_15 = 7
    MINUTES_20 = 8
    MINUTES_30 = 9
    HOURS_1 = 10
    HOURS_2 = 11
    HOURS_3 = 12
    HOURS_6 = 13
    HOURS_12 = 14
    HOURS_24 = 15

    @classmethod
    def nearest(cls, seconds: int) -> 'WakeupIntervalOgx':
        if seconds == 0:
            return WakeupIntervalOgx.NONE
        if seconds <= 45:
            return WakeupIntervalOgx.SECONDS_30
        if seconds < 3600:
            minutes = round(seconds / 60, 0)
            if minutes < 2:
                return WakeupIntervalOgx.MINUTES_1
            if minutes < 3:
                return WakeupIntervalOgx.MINUTES_2
            if minutes < 5:
                return WakeupIntervalOgx.MINUTES_3
            if minutes < 10:
                return WakeupIntervalOgx.MINUTES_5
            if minutes < 20:
                return WakeupIntervalOgx.MINUTES_15
            if minutes < 30:
                return WakeupIntervalOgx.MINUTES_20
            if minutes < 45:
                return WakeupIntervalOgx.MINUTES_30
            if minutes < 90:
                return WakeupIntervalOgx.HOURS_1
        hours = round(seconds / 3600, 0)
        if hours < 2:
            return WakeupIntervalOgx.HOURS_1
        if hours < 3:
            return WakeupIntervalOgx.HOURS_2
        if hours <= 5:
            return WakeupIntervalOgx.HOURS_3
        if hours <= 9:
            return WakeupIntervalOgx.HOURS_6
        if hours <= 18:
            return WakeupIntervalOgx.HOURS_12
        return WakeupIntervalOgx.HOURS_24


class PowerMode(IntEnum):
    """The modem power mode.
    
    Implies various internal state machine settings for balancing power
    consumption against speed of recovery from line of sight blockages.
    """
    MOBILE_POWERED = 0
    FIXED_POWERED = 1
    MOBILE_BATTERY = 2
    FIXED_BATTERY = 3
    MOBILE_MINIMAL = 4
    MOBILE_PARKED = 5


class EventNotification(IntFlag):
    """Bitmask enumerated values for modem event notifications."""
    GNSS_FIX_NEW =              0b000000000000001
    MESSAGE_MT_RECEIVED =       0b000000000000010
    MESSAGE_MO_COMPLETE =       0b000000000000100
    NETWORK_REGISTERED =        0b000000000001000
    MODEM_RESET_COMPLETE =      0b000000000010000
    JAMMING_ANTENNA_CHANGE =    0b000000000100000
    MODEM_RESET_PENDING =       0b000000001000000
    WAKEUP_INTERVAL_CHANGE =    0b000000010000000
    UTC_TIME_SYNC =             0b000000100000000
    GNSS_FIX_TIMEOUT =          0b000001000000000
    EVENT_TRACE_CACHED =        0b000010000000000
    NETWORK_PING_ACKNOWLEDGED = 0b000100000000000   # IDP-only
    MESSAGE_MO_STARTED =        0b001000000000000   # OGx-only
    SATCOM_STATE_CHANGE =       0b010000000000000   # OGx-only
    NETINFO_UPDATE =            0b100000000000000   # OGx-only
    
    @classmethod
    def get_events(cls, event_mask: int) -> 'list[EventNotification]':
        """Parses a bitmask to return a list of events."""
        return [item for item in cls if item.value & event_mask]
    
    @classmethod
    def get_bitmask(cls, events: 'list[EventNotification]') -> int:
        """Parse a list of events and return a bitmask."""
        if (not isinstance(events, list) or
            not all(isinstance(e, EventNotification) for e in events)):
            raise ValueError('Invalid list of events')
        bitmask = 0
        for event in events:
            bitmask |= event.value
        return bitmask


class SignalQuality(IntEnum):
    """A qualitative indicator of relative signal strength."""
    UNKNOWN = 0   # No Rx/Tx attempt
    NONE = 1   # T415 OGx No successful Rx/Tx
    POOR = 2   # T415 OGx Very poor
    MARGINAL = 3   # T415 OGx Poor
    FAIR = 4
    GOOD = 5
    STRONG = 6   # T415 OGx Very Good
    BEST = 7   # T415 OGx Excellent
    
    def bars(self) -> str:
        """Qualitative indicator as number of bars 0..5"""
        if self.value == 1:
            return 'BARS_0'
        elif self.value == 2:
            return 'BARS_1'
        elif self.value == 3:
            return 'BARS_2'
        elif self.value == 4:
            return 'BARS_3'
        elif self.value == 5:
            return 'BARS_4'
        elif self.value == 6:
            return 'BARS_5'
        elif self.value == 7:
            return 'BARS_6'
        return 'UNKNOWN'


class SignalLevelIdp(Enum):
    """Qualitative mapping of SNR/CN0 values for a IDP Regional Beam.
    
    NONE, FAIR, GOOD: a scale to be used as greaterOrEqual threshold
    """
    NONE = 0.0
    POOR = 37.0
    MARGINAL = 39.0
    FAIR = 41.0
    GOOD = 43.0
    STRONG = 45.5
    INVALID = 55.0
    
    @classmethod
    def nearest(cls, value: Union[float, int]) -> 'SignalLevelIdp':
        if not isinstance(value, (float, int)):
            raise ValueError('Value must be float or int')
        for member in sorted(cls, key=lambda e: e.value, reverse=True):
            if value >= member.value:
                return member
        return SignalLevelIdp.INVALID


class GeoBeam(IntEnum):
    """Geographic Beam identifiers mapped to readable names."""
    GLOBAL_BB = 0
    AMER_RB1 = 1
    AMER_RB2 = 2
    AMER_RB3 = 3
    AMER_RB4 = 4
    AMER_RB5 = 5
    AMER_RB6 = 6
    AMER_RB7 = 7
    AMER_RB8 = 8
    AMER_RB9 = 9
    AMER_RB10 = 10
    AMER_RB11 = 11
    AMER_RB12 = 12
    AMER_RB13 = 13
    AMER_RB14 = 14
    AMER_RB15 = 15
    AMER_RB16 = 16
    AMER_RB17 = 17
    AMER_RB18 = 18
    AMER_RB19 = 19
    AORW_SC = 61
    EMEA_RB1 = 21
    EMEA_RB2 = 22
    EMEA_RB3 = 23
    EMEA_RB4 = 24
    EMEA_RB5 = 25
    EMEA_RB6 = 26
    EMEA_RB7 = 27
    EMEA_RB8 = 28
    EMEA_RB9 = 29
    EMEA_RB10 = 30
    EMEA_RB11 = 31
    EMEA_RB12 = 32
    EMEA_RB13 = 33
    EMEA_RB14 = 34
    EMEA_RB15 = 35
    EMEA_RB16 = 36
    EMEA_RB17 = 37
    EMEA_RB18 = 38
    EMEA_RB19 = 39
    APAC_RB1 = 41
    APAC_RB2 = 42
    APAC_RB3 = 43
    APAC_RB4 = 44
    APAC_RB5 = 45
    APAC_RB6 = 46
    APAC_RB7 = 47
    APAC_RB8 = 48
    APAC_RB9 = 49
    APAC_RB10 = 50
    APAC_RB11 = 51
    APAC_RB12 = 52
    APAC_RB13 = 53
    APAC_RB14 = 54
    APAC_RB15 = 55
    APAC_RB16 = 56
    APAC_RB17 = 57
    APAC_RB18 = 58
    APAC_RB19 = 59
    IOE_RB1 = 101
    IOE_RB2 = 102
    IOE_RB3 = 103
    IOE_RB4 = 104
    IOE_RB5 = 105
    IOE_RB6 = 106
    IOE_RB7 = 107
    IOE_RB8 = 108
    IOE_RB9 = 109
    IOE_RB10 = 110
    IOE_RB11 = 111
    IOE_RB12 = 112
    IOE_RB13 = 113
    IOE_RB14 = 114
    IOE_RB15 = 115
    IOE_RB16 = 116
    IOE_RB17 = 117
    IOE_RB18 = 118
    IOE_RB19 = 119

    @property
    def satellite(self):
        return self.name.split('_')[0]

    @property
    def beam(self):
        return self.name.split('_')[1]
    
    @property
    def id(self):
        return self.value


class SatelliteId(IntEnum):
    """Enumerated satellite ID used by %REGINFO."""
    EMEA = 0
    AMER = 1
    APAC = 2
    IOE = 3
    AORW = 4


@dataclass
class NetInfo:
    """Key information about network acquisition."""
    network: Optional[NetworkProtocol] = None
    state: Optional[NetworkState] = None
    beam_state: Optional[BeamStateIdp] = None
    registered: bool = False
    signal_quality: SignalQuality = SignalQuality.UNKNOWN
    signal_level: Optional[float] = None
    beam_type: Optional[BeamType] = None
    geo_beam: Optional[GeoBeam] = None
    
    def to_str(self) -> str:
        obj = {k: v for k, v in vars(self).items()}
        for k, v in obj.items():
            if isinstance(v, Enum):
                obj[k] = v.name
            elif isinstance(v, float):
                obj[k] = round(v, 1)
        return json.dumps(obj)



"""Message class definitions for IoT Nano modems.
"""

from typing import Optional, Union

from .common import MessagePriorityIdp, MessageState, MessageTypeOgx, ServiceClassOgx

__all__ = ['IotNanoMessage', 'MoMessage', 'MtMessage']


class IotNanoMessage:
    """A base class for IoT Nano messages."""
    def __init__(self, name: Optional[str] = None, **kwargs) -> None:
        """Instantiates a message.
        
        Args:
            name (str): The name/handle of the message.
            **state (MessageState): The message state.
            **payload (bytes): The message payload (including SIN/MIN bytes).
            **size (int): The message size in bytes.
            **ack_bytes (int): The progress (number of bytes acknowledged).
            **priority (MessagePriority): The (IDP) message priority.
            **service_class (ServiceClass): The (OGx) service class.
            **lifetime (int): The lifetime in minutes.
        """
        self._id: Optional[str] = name
        self._state: Optional[MessageState] = None
        self._payload: Optional[bytes] = None
        self._payload_crc32: Optional[str] = None
        self._size: Optional[int] = None
        self._ack_bytes: Optional[int] = None
        self._priority: Optional[MessagePriorityIdp] = None
        self._type: Optional[MessageTypeOgx] = None
        self._service_class: Optional[ServiceClassOgx] = None
        self._lifetime: Optional[int] = None
        for k, v in kwargs.items():
            if hasattr(self, k):
                setattr(self, k, v)
    
    @property
    def id(self) -> Union[str, None]:
        return self._id
    
    @id.setter
    def id(self, value: str):
        if not isinstance(value, str) or not value:
            raise ValueError('Name must be a non-empty string')
        self._id = value
        
    @property
    def state(self) -> Union[MessageState, None]:
        return self._state
    
    @state.setter
    def state(self, value: MessageState):
        if not isinstance(value, MessageState):
            raise ValueError('Invalid message state')
        self._state = value
    
    @property
    def payload(self) -> Union[bytes, None]:
        return self._payload
    
    @payload.setter
    def payload(self, data: bytes):
        if not isinstance(data, bytes) or len(data) < 2:
            raise ValueError('Payload must be bytes with length 2 or more')
        self._payload = data
        self._size = len(data)
    
    @property
    def payload_crc32(self) -> Union[str, None]:
        return self._payload_crc32
    
    @payload_crc32.setter
    def payload_crc32(self, value: str):
        if isinstance(value, str) and len(value) == 8:
            try:
                int(value, 16)
                self._payload_crc32 = value
                return
            except ValueError:
                pass
        raise ValueError('Invalid CRC32 hex string')
    
    @property
    def codec_sin(self) -> Union[int, None]:
        return self._payload[0] if self._payload else None
    
    @property
    def codec_min(self) -> Union[int, None]:
        return self._payload[1] if self._payload else None

    @property
    def size(self) -> Union[int, None]:
        return self._size
    
    @size.setter
    def size(self, value: int):
        if not isinstance(value, int) or value < 2:
            raise ValueError('Invalid size must be 2 or more')
        if self._payload and value != len(self._payload):
            raise ValueError('Size must match payload length')
        self._size = value
    
    @property
    def ack_bytes(self) -> Union[int, None]:
        return self._ack_bytes
    
    @ack_bytes.setter
    def ack_bytes(self, value: int):
        if not isinstance(value, int) or self.size and value > self.size:
            raise ValueError('Invalid delivered must be integer up to size')
        self._ack_bytes = value
    
    @property
    def priority(self) -> Union[MessagePriorityIdp, None]:
        return self._priority
    
    @priority.setter
    def priority(self, value: MessagePriorityIdp):
        if not isinstance(value, MessagePriorityIdp):
            raise ValueError('Invalid priority')
        self._priority = value
    
    @property
    def service_class(self) -> Union[ServiceClassOgx, None]:
        return self._service_class
    
    @service_class.setter
    def service_class(self, value: ServiceClassOgx):
        if not isinstance(value, ServiceClassOgx):
            raise ValueError('Invalid service class')
        self._service_class = value
    
    @property
    def lifetime(self) -> Union[int, None]:
        """Remaining lifetime in (Tx) queue."""
        return self._lifetime
    
    @lifetime.setter
    def lifetime(self, value: int):
        if not isinstance(value, int) or value not in range(65536):
            raise ValueError('Invalid lifetime must be 0..65535 minutes')
        self._lifetime = value
    
    @property
    def type(self) -> Union[MessageTypeOgx, None]:
        return self._type
    
    @type.setter
    def type(self, value: MessageTypeOgx):
        if not isinstance(value, MessageTypeOgx):
            raise ValueError('Invalid service class')
        self._type = value


class MoMessage(IotNanoMessage):
    """A Mobile-Originated message."""


class MtMessage(IotNanoMessage):
    """A Mobile-Terminated message."""
    @property
    def lifetime(self) -> Union[int, None]:
        return self._lifetime
    
    @lifetime.setter
    def lifetime(self, value: int) -> None:
        raise ValueError('MT message does not have lifetime')



"""Classes and methods for location, elevation and azimuth for satellite modems.

Parses NMEA-0183 data into a `GnssLocation` object.
"""

import inspect
import json
import logging
from dataclasses import dataclass
from enum import IntEnum

from .utils import iso_to_ts, ts_to_iso


_log = logging.getLogger(__name__)


class GnssFixType(IntEnum):
    """Enumerated fix type from NMEA-0183 standard."""
    NONE = 1
    FIX_2D = 2
    FIX_3D = 3


class GnssFixQuality(IntEnum):
    """Enumerated fix quality from NMEA-0183 standard."""
    INVALID = 0
    GPS_SPS = 1
    DGPS = 2
    PPS = 3
    RTK = 4
    FLOAT_RTK = 5
    EST_DEAD_RECKONING = 6
    MANUAL = 7
    SIMULATION = 8


@dataclass
class GnssSatelliteInfo(object):
    """Information specific to a GNSS satellite.
    
    Attributes:
        prn: The PRN code (Pseudo-Random Number sequence)
        elevation: The satellite elevation
        azimuth: The satellite azimuth
        snr: The satellite Signal-to-Noise Ratio
    """
    prn: int
    elevation: int
    azimuth: int
    snr: int


def validate_nmea(nmea_sentence: str) -> bool:
    """Validates a given NMEA-0183 sentence with CRC.
    
    Args:
        nmea_sentence (str): NMEA-0183 sentence ending in checksum.
    
    """
    if '*' not in nmea_sentence:
        return False
    data, cs_hex = nmea_sentence.split('*')
    candidate = int(cs_hex, 16)
    crc = 0   # initial
    for i in range(1, len(data)):   # ignore initial $
        crc ^= ord(data[i])
    return candidate == crc


class GnssLocation:
    """A set of location-based information derived from the modem's NMEA data.
    
    Uses 90.0/180.0 if latitude/longitude are unknown

    Attributes:
        latitude (float): decimal degrees
        longitude (float): decimal degrees
        altitude (float): in metres
        speed (float): in knots
        heading (float): in degrees
        timestamp (int): in seconds since 1970-01-01T00:00:00Z
        satellites (int): in view at time of fix
        fix_type (GnssFixType): 1=None, 2=2D or 3=3D
        fix_quality (GnssFixQuality): Enumerated lookup value
        pdop (float): Probability Dilution of Precision
        hdop (float): Horizontal Dilution of Precision
        vdop (float): Vertical Dilution of Precision
        time_iso (str): ISO 8601 formatted timestamp

    """
    __slots__ = ('latitude', 'longitude', 'altitude', 'speed', 'heading',
                 'timestamp', 'satellites', 'fix_type', 'fix_quality',
                 'pdop', 'hdop', 'vdop',)
    
    def __init__(self, **kwargs):
        """Initializes a Location with default latitude/longitude 90/180."""
        self.latitude = float(kwargs.get('latitude', 90.0))
        self.longitude = float(kwargs.get('longitude', 180.0))
        self.altitude = float(kwargs.get('altitude', 0.0))   # metres
        self.speed = float(kwargs.get('speed', 0.0))  # knots
        self.heading = float(kwargs.get('heading', 0.0))   # degrees
        self.timestamp = int(kwargs.get('timestamp', 0))   # seconds (unix)
        self.satellites = int(kwargs.get('satellites', 0))
        self.fix_type = GnssFixType(int(kwargs.get('fix_type', 1)))
        self.fix_quality = GnssFixQuality(int(kwargs.get('fix_quality', 0)))
        self.pdop = float(kwargs.get('pdop', 99))
        self.hdop = float(kwargs.get('hdop', 99))
        self.vdop = float(kwargs.get('vdop', 99))

    @property
    def time_iso(self) -> str:
        return f'{ts_to_iso(self.timestamp)}'

    def __repr__(self) -> str:
        obj = {k: v for k, v in vars(self).items()
               if not k.startswith('_') and not callable(v)}
        for name, _ in inspect.getmembers(self.__class__,
                                          lambda o: isinstance(o, property)):
            if not name.startswith('_'):
                try:
                    v = getattr(self, name)
                    if not callable(v):
                        obj[name] = v
                except Exception:
                    pass
        for k, v in obj.items():
            if k in ['latitude', 'longitude']:
                obj[k] = round(v, 5)
            elif isinstance(v, float):
                obj[k] = round(v, 1)
            elif isinstance(v, IntEnum):
                obj[k] = v.name
        return json.dumps(obj, skipkeys=True)
    
    def is_valid(self) -> bool:
        """Check validity."""
        return self.latitude is not None and self.longitude is not None
    
    def parse_nmea(self, nmea_sentence: str):
        """Update the location with information derived from an NMEA sentence.
        
        Args:
            nmea_sentence (str): The NMEA-0183 sentence to parse.
        """
        if not validate_nmea(nmea_sentence):
            raise ValueError('Invalid NMEA-0183 sentence')
        data = nmea_sentence.split('*')[0]
        nmea_type = ''
        cache = {}
        for i, field_data in enumerate(data.split(',')):
            if i == 0:
                nmea_type = field_data[-3:]
                if nmea_type == 'GSV':
                    _log.warning('No processing required for GSV sentence')
                    return
                if nmea_type == 'GSA' and self.vdop != 99:
                    # _log.debug('Skipping redundant GSA data')
                    return
                # _log.debug('Processing NMEA type: %s', nmea_type)
            elif i == 1:
                if nmea_type == 'RMC':
                    cache['fix_hour'] = field_data[0:2]
                    cache['fix_min'] = field_data[2:4]
                    cache['fix_sec'] = field_data[4:6]
            elif i == 2:
                if nmea_type == 'RMC':
                    if (field_data == 'V'):
                        _log.warn('Fix Void')
                elif nmea_type == 'GSA':
                    self.fix_type = GnssFixType(int(field_data))
            elif i == 3:
                if nmea_type == 'RMC':
                    self.latitude = (float(field_data[0:2]) +
                                     float(field_data[2:]) / 60.0)
            elif i == 4:
                if nmea_type == 'RMC':
                    if field_data == 'S':
                        self.latitude *= -1
            elif i == 5:
                if nmea_type == 'RMC':
                    self.longitude = (float(field_data[0:3]) +
                                      float(field_data[3:]) / 60.0)
            elif i == 6:
                if nmea_type == 'RMC':
                    if field_data == 'W':
                        self.longitude *= -1
                elif nmea_type == 'GGA':
                    self.fix_quality = GnssFixQuality(int(field_data))
            elif i == 7:
                if nmea_type == 'RMC':
                    self.speed = float(field_data)
                elif nmea_type == 'GGA':
                    self.satellites = int(field_data)
            elif i == 8:
                if nmea_type == 'RMC':
                    self.heading = float(field_data)
                elif nmea_type == 'GGA':
                    self.hdop = round(float(field_data), 1)
            elif i == 9:
                if nmea_type == 'RMC':
                    fix_day = field_data[0:2]
                    fix_month = field_data[2:4]
                    fix_yy = int(field_data[4:])
                    fix_yy += 1900 if fix_yy >= 73 else 2000
                    iso_time = (f'{fix_yy}-{fix_month}-{fix_day}T'
                                f'{cache["fix_hour"]}:{cache["fix_min"]}'
                                f':{cache["fix_sec"]}Z')
                    self.timestamp = int(iso_to_ts(iso_time))
                elif nmea_type == 'GGA':
                    self.altitude = float(field_data)
            elif i == 10:
                # RMC magnetic variation - ignore
                if nmea_type == 'GGA' and field_data != 'M':
                    _log.warning('Unexpected altitude units: %s', field_data)
            # elif i == 11:   # RMC magnetic variation direction, GGA height of geoid - ignore
            # elif i == 12:   # GGA units height of geoid - ignore
            # elif i == 13:   # GGA seconds since last DGPS update - ignore
            # elif i == 14:   # GGA DGPS station ID - ignore
            elif i == 15:   # GSA PDOP - ignore (unused)
                if nmea_type == 'GSA':
                    self.pdop = round(float(field_data), 1)
            # elif i == 16:   # GSA HDOP - ignore (use GGA)
            elif i == 17:
                if nmea_type == 'GSA':
                    self.vdop = round(float(field_data), 1)
    
    @classmethod
    def from_nmea_list(cls, nmea_list: list[str]) -> 'GnssLocation':
        """Create a GnssLocation from a list of NMEA-0183 sentences."""
        if (not isinstance(nmea_list, list) or
            not all(isinstance(s, str) for s in nmea_list)):
            raise ValueError('Invalid list or sentences in list')
        loc = GnssLocation()
        for sentence in nmea_list:
            loc.parse_nmea(sentence)
        return loc
